/*
* Copyright (C) 2017
* Computer Graphics Group, The Institute of Computer Graphics and Algorithms, TU Wien
* Written by Tobias Klein <tklein@cg.tuwien.ac.at>
* All rights reserved.
*/

#include "GLWidget.h"

#include "glsw.h"
#include <glm/gtc/type_ptr.hpp>
#include <glm/gtc/matrix_transform.hpp>
#include <QMouseEvent>
#include <QDir>
#include <qopenglwidget.h>
#include <qmessagebox.h>
#include <qopenglframebufferobject.h>
#include <QOpenGLTexture>
#include <ctime>
#include <ppl.h>
#include <sstream>
#include <iostream>
#include <fstream>

#include "MainWindow.h"

#define GL_GPU_MEM_INFO_TOTAL_AVAILABLE_MEM_NVX 0x9048
#define GL_GPU_MEM_INFO_CURRENT_AVAILABLE_MEM_NVX 0x9049

GLfloat cube_vertices[] = {
	// front
	-1.0, -1.0,  1.0,
	1.0, -1.0,  1.0,
	1.0,  1.0,  1.0,
	-1.0,  1.0,  1.0,
	// back
	-1.0, -1.0, -1.0,
	1.0, -1.0, -1.0,
	1.0,  1.0, -1.0,
	-1.0,  1.0, -1.0,
};

GLfloat tex_coordinates[] = {
	// front
	0.0, 0.0,  1.0,
	1.0, 0.0,  1.0,
	1.0,  1.0,  1.0,
	0.0,  1.0,  1.0,
	// back
	0.0, 0.0, 0.0,
	1.0, 0.0, 0.0,
	1.0,  1.0, 0.0,
	0.0,  1.0, 0.0,
};

GLushort cube_elements[] = {
	// front
	0, 1, 2,
	2, 3, 0,
	// top
	1, 5, 6,
	6, 2, 1,
	// back
	7, 6, 5,
	5, 4, 7,
	// bottom
	4, 0, 3,
	3, 7, 4,
	// left
	4, 5, 1,
	1, 0, 4,
	// right
	3, 2, 6,
	6, 7, 3,
};

GLfloat quad_vertices[] =
{
	-1.0f, -1.0f, 0.0f,
	1.0f, -1.0f, 0.0f,
	1.0f, 1.0f, 0.0f,
	-1.0f, 1.0f, 0.0f
};

GLushort quad_elements[] =
{
	0, 1, 2, 
	0, 2, 3
};

const QString assetsPath = QDir::currentPath() + "/../assets/";

const QString shaderPath1 = assetsPath + "shader/cube.glsl";
const QString shaderPath2 = assetsPath + "shader/raycasting.glsl";



GLWidget::GLWidget(QWidget *parent, MainWindow *mainWindow)
	: QOpenGLWidget(parent),
	m_vertexBuffer(QOpenGLBuffer::VertexBuffer),
	m_indexBuffer(QOpenGLBuffer::IndexBuffer),
	m_vertexBufferQuad(QOpenGLBuffer::VertexBuffer),
	m_indexBufferQuad(QOpenGLBuffer::IndexBuffer)
{
	m_MainWindow = mainWindow;
	m_renderingMode = 3;
	QSurfaceFormat format;
	// asks for a OpenGL 3.2 debug context using the Core profile
	format.setMajorVersion(3);
	format.setMinorVersion(2);
	format.setProfile(QSurfaceFormat::CoreProfile);
	format.setOption(QSurfaceFormat::DebugContext);

	setFormat(format);
	
	// we use a file watcher to recompile shaders on the fly
	// you can change shaders and directly see the result while the program is running
	m_fileWatcher = new QFileSystemWatcher(this);
	connect(m_fileWatcher, SIGNAL(fileChanged(const QString &)), this, SLOT(fileChanged(const QString &)));
	m_fileWatcher->addPath(QDir::currentPath() + shaderPath1);
	m_fileWatcher->addPath(QDir::currentPath() + shaderPath2);

	initglsw(); // initializes the OpenGL Shader Wrangler
}


GLWidget::~GLWidget()
{
	glswShutdown(); // shuts down the OpenGL Shader Wrangler
}

// initializes the OpenGL Shader Wrangler
void GLWidget::initglsw()
{
	glswInit();
	QString shader_folder = assetsPath + "shader/";
	QByteArray ba = shader_folder.toLatin1();
	const char *c_str = ba.data();
	glswSetPath(c_str, ".glsl");
	qDebug() << "glsw path: " << c_str;
	glswAddDirectiveToken("", "#version 330");
}

// the rendering routing, this function is called for every frame
void GLWidget::paintGL()
{ 
	// you must transpose the matrix when it is passed from glm to QT!
	QMatrix4x4 modelView = QMatrix4x4(glm::value_ptr(m_camera.getViewMatrix())).transposed();
	QMatrix4x4 projection = QMatrix4x4(glm::value_ptr(m_camera.getProjectionMatrix())).transposed();

	QOpenGLVertexArrayObject::Binder vaoBinder(&m_vaoCube);
	m_programCube->bind();

	m_programCube->setUniformValue(m_programCube->uniformLocation("mvMatrix"), modelView);
	m_programCube->setUniformValue(m_programCube->uniformLocation("projMatrix"), projection);

	m_programCube->setUniformValue("scaleX", scaleX);
	m_programCube->setUniformValue("scaleY", scaleY);
	m_programCube->setUniformValue("scaleZ", scaleZ);

	// 1. render front faces to FBO
	glCullFace(GL_BACK);
	m_FBO_frontFaces->bind();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawElements(GL_TRIANGLES, sizeof(cube_elements) / sizeof(cube_elements[0]), GL_UNSIGNED_SHORT, 0);
	
	m_FBO_frontFaces->release();


	// 2. render back faces to FBO
	glCullFace(GL_FRONT);
	m_FBO_backFaces->bind();

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	glDrawElements(GL_TRIANGLES, sizeof(cube_elements) / sizeof(cube_elements[0]), GL_UNSIGNED_SHORT, 0);

	m_FBO_backFaces->release();


	vaoBinder.release();

	glCullFace(GL_BACK);
	// 3. render the volume

	glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

	QOpenGLVertexArrayObject::Binder vaoBinder2(&m_vaoQuad);
	m_programVolume->bind();

	m_programVolume->setUniformValue("renderingMode", m_renderingMode);

	glActiveTexture(GL_TEXTURE0);
	glBindTexture(GL_TEXTURE_2D, m_FBO_frontFaces->texture());
	m_programVolume->setUniformValue("frontFaces", 0);

	glActiveTexture(GL_TEXTURE0 + 1);
	glBindTexture(GL_TEXTURE_2D, m_FBO_backFaces->texture());
	m_programVolume->setUniformValue("backFaces", 1);




	if (sideBySide) {
		glViewport(0, 0, width() / 2, height());		
	}
	else {
		glViewport(0, 0, width(), height());
	}


	glActiveTexture(GL_TEXTURE0 + 2);
	glBindTexture(GL_TEXTURE_3D, m_VolumeTexture->textureId());
	m_programVolume->setUniformValue("volume", 2);

	glDrawElements(GL_TRIANGLES, sizeof(quad_elements) / sizeof(quad_elements[0]), GL_UNSIGNED_SHORT, 0);

	if (sideBySide) {
		glViewport(width() / 2, 0, width() / 2, height());

		glActiveTexture(GL_TEXTURE0 + 2);
		glBindTexture(GL_TEXTURE_3D, m_VolumeTexture2->textureId());
		m_programVolume->setUniformValue("volume", 2);


		glDrawElements(GL_TRIANGLES, sizeof(quad_elements) / sizeof(quad_elements[0]), GL_UNSIGNED_SHORT, 0);
	}
	vaoBinder2.release();
}

void GLWidget::initializeGL()
{
	connect(context(), &QOpenGLContext::aboutToBeDestroyed, this, &GLWidget::cleanup);

	m_logger = new QOpenGLDebugLogger(this);
	m_logger->initialize();
	connect(m_logger, &QOpenGLDebugLogger::messageLogged, this, &GLWidget::handleLoggedMessage, Qt::DirectConnection);

	QWidget::setFocusPolicy(Qt::FocusPolicy::ClickFocus);

	m_VolumeTexture = new QOpenGLTexture(QOpenGLTexture::Target3D);
	m_VolumeTexture->setMinMagFilters(QOpenGLTexture::Nearest, QOpenGLTexture::Nearest);
	m_VolumeTexture->setFormat(QOpenGLTexture::RGB32F);

	m_VolumeTexture2 = new QOpenGLTexture(QOpenGLTexture::Target3D);
	m_VolumeTexture2->setMinMagFilters(QOpenGLTexture::Nearest, QOpenGLTexture::Nearest);
	m_VolumeTexture2->setFormat(QOpenGLTexture::RGB32F);


	initializeOpenGLFunctions();
	glClearColor(0.0f, 0.0f, 0.0f, 1.0f);
	glEnable(GL_DEPTH_TEST);
	glEnable(GL_CULL_FACE);
	glCullFace(GL_FRONT);


	// create programs
	m_programCube = new QOpenGLShaderProgram();
	m_programVolume = new QOpenGLShaderProgram();

	loadShaders();
	
	// create Frame Buffer Object (FBO)
	m_FBO_frontFaces = new QOpenGLFramebufferObject(width(), height());
	m_FBO_backFaces = new QOpenGLFramebufferObject(width(), height());

	if (!m_vaoCube.create()) {
		qDebug() << "error creating vao";
	}

	// bind vertex array object (VAO)
	// the VAO encapsulates all the data that is associated with the vertex processor
	// see: http://ogldev.atspace.co.uk/www/tutorial32/tutorial32.html
	QOpenGLVertexArrayObject::Binder vaoBinder(&m_vaoCube);
	
	// Setup our vertex buffer object
	m_vertexBuffer.create();
	m_vertexBuffer.bind();
	// Setup our index buffer object
	m_indexBuffer.create();
	m_indexBuffer.bind();

	// allocate vertices and indices of a cube
	m_vertexBuffer.allocate(cube_vertices, 8 * 3 * sizeof(GLfloat));
	m_indexBuffer.allocate(cube_elements, 12 * 3 * sizeof(GLushort)); 

	m_programCube->enableAttributeArray(0);
	m_programCube->enableAttributeArray(1);

	m_programCube->setAttributeBuffer(
		0,			// Specifies the index of the generic vertex attribute to be modified
		GL_FLOAT,	// Specifies the data type of each component in the array
		0,			// Specifies the byte offset between consecutive generic vertex attributes
		3			// Specifies the number of components per generic vertex attribute
	);


	// Setup our tex coord buffer object
	m_texCoordBuffer.create();
	m_texCoordBuffer.bind();

	m_texCoordBuffer.allocate(tex_coordinates, 8 * 3 * sizeof(GLfloat));
	
	m_programCube->setAttributeBuffer(
		1,			// Specifies the index of the generic vertex attribute to be modified
		GL_FLOAT,	// Specifies the data type of each component in the array
		0,			// Specifies the byte offset between consecutive generic vertex attributes
		3			// Specifies the number of components per generic vertex attribute
	);

	if (!m_vaoQuad.create()) {
		qDebug() << "error creating vao";
	}

	QOpenGLVertexArrayObject::Binder vaoBinder2(&m_vaoQuad);
	
	m_vertexBufferQuad.create();
	m_vertexBufferQuad.bind();
	
	// Setup our index buffer object
	m_indexBufferQuad.create();
	m_indexBufferQuad.bind();

	m_vertexBufferQuad.allocate(quad_vertices, 4 * 3 * sizeof(GLfloat));
	m_indexBufferQuad.allocate(quad_elements, 2 * 3 * sizeof(GLushort));

	m_programVolume->setAttributeBuffer(
		0,			// Specifies the index of the generic vertex attribute to be modified
		GL_FLOAT,	// Specifies the data type of each component in the array
		0,			// Specifies the byte offset between consecutive generic vertex attributes
		3			// Specifies the number of components per generic vertex attribute
	);

	m_programVolume->enableAttributeArray(0);
	
	GLint total_mem_kb = 0;
	glGetIntegerv(GL_GPU_MEM_INFO_TOTAL_AVAILABLE_MEM_NVX,
		&total_mem_kb);

	GLint cur_avail_mem_kb = 0;
	glGetIntegerv(GL_GPU_MEM_INFO_CURRENT_AVAILABLE_MEM_NVX,
		&cur_avail_mem_kb);

	float cur_avail_mem_mb = float(cur_avail_mem_kb) / 1024.0f;
	float total_mem_mb = float(total_mem_kb) / 1024.0f;

	m_MainWindow->displayTotalGPUMemory(total_mem_mb);
	m_MainWindow->displayUsedGPUMemory(total_mem_mb - cur_avail_mem_mb);
	
	generateDirs();

	m_camera.setAspect(float(width()) / float(height()));
}

void GLWidget::generateDirs() {

	dir_x.clear();

	std::vector<glm::mat4> XY_rot;
	std::vector<glm::mat4> XZ_rot;

	for (int i = 0; i < 360; i += _delta) {
		XY_rot.push_back(glm::rotate(glm::mat4(), (float)i, glm::vec3(0, 0, 1)));
		XZ_rot.push_back(glm::rotate(glm::mat4(), (float)i, glm::vec3(0, 1, 0)));
	}

	auto start = glm::vec4(1, 0, 0, 1);

	for (auto &xy : XY_rot) {
		for (auto &xz : XZ_rot) {
			auto v = xy * xz * start;
			if (v.y >= 0) {
				/*
				"Since both vectors x and -x could define
				the same line, we consistently select x with a positive
				y-coordinate."
				*/
				if (v.w != 1) {
					std::cout << "Err: dir_x generation, invalid w value" << std::endl;
				}

				//normalize just in case...
				dir_x.push_back(glm::normalize(glm::vec3(v)));
			}
		}
	}

}

void GLWidget::filterChanged(const QString & text)
{
	if (unfiltered == nullptr) {
		std::cout << "Error: Load volume first" << std::endl;
		return;
	}

	if (text != nullptr) {
		filter = text.toStdString();
	}

	if (filter == "Streamline") {
		if (streamline == nullptr) {
			delete streamline;
			streamline = streamlineAlgorithm();
		}
		destroyRecreateVolTexture(streamline);
	}
	else if (filter == "Avg Gauss") {

		if (gauss == nullptr) {
			calcGauss();
		}
		destroyRecreateVolTexture(gauss);
	}
	else {
		destroyRecreateVolTexture(unfiltered);
	}

}

void GLWidget::calcGauss() {
	std::cout << "Gauss Filtering" << std::endl;

	Volume* newVol = new Volume(unfiltered->width(), unfiltered->height(), unfiltered->depth());

	std::map<int, std::map<int, std::map<int, float>>> gaussKernel;

	gaussKernel[-1][-1][-1] = 1.0f / 64.0f;
	gaussKernel[-1][-1][0] = 2.0f / 64.0f;
	gaussKernel[-1][-1][1] = 1.0f / 64.0f;

	gaussKernel[-1][0][-1] = 2.0f / 64.0f;
	gaussKernel[-1][0][0] = 4.0f / 64.0f;
	gaussKernel[-1][0][1] = 2.0f / 64.0f;

	gaussKernel[-1][1][-1] = 1.0f / 64.0f;
	gaussKernel[-1][1][0] = 2.0f / 64.0f;
	gaussKernel[-1][1][1] = 1.0f / 64.0f;



	gaussKernel[0][-1][-1] = 2.0f / 64.0f;
	gaussKernel[0][-1][0] = 4.0f / 64.0f;
	gaussKernel[0][-1][1] = 2.0f / 64.0f;
				
	gaussKernel[0][0][-1] = 4.0f / 64.0f;
	gaussKernel[0][0][0] = 8.0f / 64.0f;
	gaussKernel[0][0][1] = 4.0f / 64.0f;
				
	gaussKernel[0][1][-1] = 2.0f / 64.0f;
	gaussKernel[0][1][0] = 4.0f / 64.0f;
	gaussKernel[0][1][1] = 2.0f / 64.0f;



	gaussKernel[1][-1][-1] = 1.0f / 64.0f;
	gaussKernel[1][-1][0] = 2.0f / 64.0f;
	gaussKernel[1][-1][1] = 1.0f / 64.0f;

	gaussKernel[1][0][-1] = 2.0f / 64.0f;
	gaussKernel[1][0][0] = 4.0f / 64.0f;
	gaussKernel[1][0][1] = 2.0f / 64.0f;

	gaussKernel[1][1][-1] = 1.0f / 64.0f;
	gaussKernel[1][1][0] = 2.0f / 64.0f;
	gaussKernel[1][1][1] = 1.0f / 64.0f;


	concurrency::parallel_for(int(0), unfiltered->width(), [&](int xp) {

	//for(int xp = 0; xp < unfiltered->width(); xp++) {
		for (int yp = 0; yp < unfiltered->height(); yp++) {
			for (int zp = 0; zp < unfiltered->depth(); zp++) {

				bool skip = false;
				float newVal = 0;

				for (int x = -1; x <= 1 && !skip; x++) {
					for (int y = -1; y <= 1 && !skip; y++) {
						for (int z = -1; z <= 1 && !skip; z++) {

							int newX = xp + x;
							int newY = yp + y;
							int newZ = zp + z;

							if (newX < 0 || newX >= unfiltered->width()
								|| newY < 0 || newY >= unfiltered->height()
								|| newZ < 0 || newZ >= unfiltered->depth()) {
								skip = true;
								newVol->set(0.0f, xp, yp, zp);
								break;
							}

							newVal += unfiltered->voxel(newX, newY, newZ).getValue() * gaussKernel[x][y][z];
						}
					}
				}

				if (!skip) {
					newVol->set(newVal, xp, yp, zp);
				}
			}
		}
	}
	);

	gauss = newVol;
}

void GLWidget::calcStreamline() {
	streamline = streamlineAlgorithm();
}

void GLWidget::destroyRecreateVolTexture(Volume* vol, int n) {

	makeCurrent();

	QOpenGLTexture* tex = new QOpenGLTexture(QOpenGLTexture::Target3D);
	tex->setMinMagFilters(QOpenGLTexture::Linear, QOpenGLTexture::Linear);
	tex->setFormat(QOpenGLTexture::RGB32F);

	tex->setSize(vol->width(), vol->height(), vol->depth());

	tex->allocateStorage(QOpenGLTexture::Red, QOpenGLTexture::Float32); // works
	tex->setData(QOpenGLTexture::Red, QOpenGLTexture::Float32, vol->voxels()); // works

	if (n == 1) {
		m_VolumeTexture->destroy();
		m_VolumeTexture = tex;
	}
	else if (n == 2) {
		m_VolumeTexture2->destroy();
		m_VolumeTexture2 = tex;
	}

	doneCurrent();
}

// this function is called when the volume texture is loaded 
void GLWidget::volumeTextureLoaded(Volume *v, std::string name)
{
	volumeName = name;
	unfiltered = v;

	delete streamline;
	streamline = nullptr;
	delete gauss;
	gauss = nullptr;

	int max = std::max(std::max(unfiltered->width(), unfiltered->height()), unfiltered->depth());

	scaleX = (float)unfiltered->width() / max;
	scaleY = (float)unfiltered->height() / max;
	scaleZ = (float)unfiltered->depth() / max;

	filterChanged();
}

VectorField* GLWidget::localDirectionOfLowestVariance() {

	std::cout << "Local Direction of Lowest Variance" << std::endl;

	auto lowestVarDirs = new VectorField(unfiltered->width(), unfiltered->height(), unfiltered->depth());

	std::string vectorfieldsPath = assetsPath.toStdString() + "vectorfields/";
	std::string file = vectorfieldsPath + volumeName + "_d" + std::to_string(_delta) + "_n" + std::to_string(_n) + ".txt";

	if (!CreateDirectory(vectorfieldsPath.c_str(), NULL) && ERROR_ALREADY_EXISTS == GetLastError())
	{
		std::ifstream input(file);

		if (input.good()) {
			//file exists - read in from file
			std::cout << "Reading vector field from file " << file << std::endl;
			read(input, lowestVarDirs, unfiltered->size());
			return lowestVarDirs;
		}
	}




	//no previously generated vector field, generate a new one and write it to file

	std::cout << "generating vector field" << std::endl;

	int setSize = 2 * _n + 1; // to cover [-n;n] range

	std::atomic<int> count = 0; //track progress

	clock_t begin = clock();

	concurrency::parallel_for(int(0), unfiltered->width(), [&](int xp) {

		// just to keep track of the progress (not 100% concurrent)
		count++;

		if (count % 10 == 0) {
			std::stringstream msg;
			msg << count << "/" << unfiltered->width() << "\n";
			std::cout << msg.str();
		}
		//

		for (int yp = 0; yp < unfiltered->height(); yp++) {
			for (int zp = 0; zp < unfiltered->depth(); zp++) {

				auto P = glm::vec3(xp, yp, zp);

				int dOfMinVar = -1;
				float minVar = std::numeric_limits<float>::max();

				for (int d = 0; d < dir_x.size(); d++) {
					bool quit = false;
					auto x = dir_x[d];

					float* varSet = new float[setSize];

					for (int k = -_n; k <= _n; k++) {

						auto v = unfiltered->voxelInterpolate(P + (float)k * x);

						if (v == -1) {
							//only take x's into account where all values are valid
							quit = true;
							break;

						}
						else {
							varSet[k + _n] = v;
						}
					}

					if (quit) {
						delete[] varSet;
						continue;
					}

					float sum = 0;
					for (int i = 0; i < setSize; i++) {
						sum += varSet[i];
					}
					auto mean = sum / setSize;

					float var = 0;
					for (int i = 0; i < setSize; i++) {
						var += pow((varSet[i] - mean), 2);
					}

					delete[] varSet;

					//var = var / (float)(setSize - 1); //division eig unnoetig

					if (var <= minVar) {
						minVar = var;
						dOfMinVar = d;
					}

				}

				lowestVarDirs->set(
					dOfMinVar < 0 ? glm::vec3(0, 0, 0) : dir_x[dOfMinVar],
					xp, yp, zp);
			}
		}
	}
	);

	clock_t end = clock();
	double elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;

	std::cout << "time: " << elapsed_secs << "s" << std::endl;

	write(file, lowestVarDirs);

	return lowestVarDirs;
}

Volume* GLWidget::streamlineAlgorithm() {

	Volume* vol;

	//check if there are already filtered volumes available

	std::string filteredFolder = assetsPath.toStdString() + "filteredvolumes/";
	std::string filteredFile = filteredFolder + volumeName + "_d" + std::to_string(_delta) + "_n" + std::to_string(_n) + "_m" + std::to_string(_m) + ".txt";


	if (!CreateDirectory(filteredFolder.c_str(), NULL))
	{
		if (ERROR_ALREADY_EXISTS == GetLastError()) {

			std::ifstream input(filteredFile);

			if (input.good()) {
				std::cout << "Reading filtered volume from file " << filteredFile << std::endl;

				vol = new Volume(unfiltered->width(), unfiltered->height(), unfiltered->depth());

				if (!readFilteredVol(input, vol, unfiltered->size())) {
					std::cout << "Err reading filtered volume" << std::endl;
				}
				else {
					return vol;
				}
			}

		}
		else
		{
			// Failed to create directory.
			std::cout << "Err: failed to create " << filteredFolder << std::endl;
		}
	}

	auto lowestVarDirs = localDirectionOfLowestVariance();
	vol = streamlineIntegrationAndFiltering(lowestVarDirs);
	return vol;
}

Volume* GLWidget::streamlineIntegrationAndFiltering(VectorField* lowestVarDirs) {

	std::cout << "Streamline Integration and Filtering" << std::endl;

	std::atomic<int> count = 0; //track progress
	Volume* newVol = new Volume(unfiltered->width(), unfiltered->height(), unfiltered->depth());

	clock_t begin = clock();

	concurrency::parallel_for(int(0), unfiltered->width(), [&](int xp) {
	//for (int xp = 0; xp < unfiltered->width(); xp++) {

		// just to keep track of the progress (not 100% concurrent)
		count++;
		
		if (count % 10 == 0) {
			std::stringstream msg;
			msg << count << "/" << unfiltered->width() << "\n";
			std::cout << msg.str();
		}
		//

		for (int yp = 0; yp < unfiltered->height(); yp++) {
			for (int zp = 0; zp < unfiltered->depth(); zp++) {

				bool invalid = false; // for invalid direction vectors (e.g. bad index)

				auto P = glm::vec3(xp, yp, zp);

				float samplesSum = unfiltered->voxel(xp, yp, zp).getValue();

				//1. Forward integration
				//http://graphics.cs.ucdavis.edu/~joy/ecs277/other-notes/Numerical-Methods-for-Particle-Tracing-in-Vector-Fields.pdf
				auto p_i = P;
				for (int i = 0; i < _m; i++) {
					auto k_1 = lowestVarDirs->interpolate(p_i);
					if (zero(k_1)) { invalid = true; break; }

					auto k_2 = lowestVarDirs->interpolate(p_i + k_1 / 2.0f);
					if (zero(k_2)) { invalid = true; break; }

					auto k_3 = lowestVarDirs->interpolate(p_i + k_2 / 2.0f);
					if (zero(k_3)) { invalid = true; break; }

					auto k_4 = lowestVarDirs->interpolate(p_i + k_3);
					if (zero(k_4)) { invalid = true; break; }

					p_i = p_i + (k_1 + 2.0f * k_2 + 2.0f * k_3 + k_4) / 6.0f;

					samplesSum += unfiltered->voxelInterpolate(p_i);
				}

				//2. Backward integration
				p_i = P; //start in the middle again
				for (int i = 0; i < _m; i++) {
					//intvert the whole vectorfield -> invert all vectors read from the field

					auto k_1 = -lowestVarDirs->interpolate(p_i);
					if (zero(k_1)) { invalid = true; break; }

					auto k_2 = -lowestVarDirs->interpolate(p_i + k_1 / 2.0f);
					if (zero(k_2)) { invalid = true; break; }

					auto k_3 = -lowestVarDirs->interpolate(p_i + k_2 / 2.0f);
					if (zero(k_3)) { invalid = true; break; }

					auto k_4 = -lowestVarDirs->interpolate(p_i + k_3);
					if (zero(k_4)) { invalid = true; break; }

					p_i = p_i + (k_1 + 2.0f * k_2 + 2.0f * k_3 + k_4) / 6.0f;

					samplesSum += unfiltered->voxelInterpolate(p_i);
				}


				//Note: possible loss of information by setting invalid values to zero, 
				//      probably just better filtering though
				float newValue = invalid ? 0 : samplesSum / (2 * _m + 1);
				newVol->set(newValue, xp, yp, zp);
			}
		}
	}
	);

	std::string file = assetsPath.toStdString() + "filteredvolumes/" + volumeName + "_d" + std::to_string(_delta) + "_n" + std::to_string(_n) + "_m" + std::to_string(_m) + ".txt";
	write(file, newVol);

	return newVol;
}

bool GLWidget::read(std::ifstream& input, VectorField* lowestVarDirs, int volSize) {

	if (input.is_open())
	{
		int i = 0;
		std::string line;

		std::vector<float[3]> vecs(volSize);

		while (getline(input, line))
		{
			std::istringstream ss(line);
			if (!(ss >> vecs[i][0] >> vecs[i][1] >> vecs[i][2]))
			{
				// error: didn't convert to a float
				std::cout << "Error reading vector field: invalid values";
				return false;
			}
			i++;
		}

		if (i != volSize) {
			std::cout << "Error reading vector field: invalid number of vectors";
			return false;
		}


		int digits = 0;
		while (pow(10, digits) < volSize) { digits++; }
		int blocksize = pow(10, (digits - 2));

		concurrency::parallel_for(int(0), volSize / blocksize, [&](int block) {

			int j = block * blocksize;

			while (j < (block + 1) * blocksize  && j < volSize) {

				lowestVarDirs->set(glm::vec3(vecs[j][0], vecs[j][1], vecs[j][2]), j);
				j++;
			}

		}
		);

		input.close();
	}
	else {
		std::cout << "Unable to open file";
		return false;
	}
	std::cout << "done reading" << std::endl;
	return true;
}

bool GLWidget::readFilteredVol(std::ifstream& input, Volume* vol, int volSize) {

	clock_t begin = clock();

	if (input.is_open())
	{
		int i = 0;
		std::string line;
		float* val = new float[volSize];

		while (getline(input, line))
		{
			std::istringstream ss(line);
			if (!(ss >> val[i++]))
			{
				// error: didn't convert to a float
				std::cout << "Error reading filtered volume: invalid values";
				return false;
			}
		}

		if (i != volSize) {
			std::cout << "Error reading vector field: invalid number of vectors";
			return false;
		}


		int digits = 0;
		while (pow(10, digits) < volSize) { digits++; }
		int blocksize = pow(10, (digits - 2));

		concurrency::parallel_for(int(0), volSize / blocksize, [&](int block) {

			int i = block * blocksize;

			while (i < (block + 1) * blocksize  && i < volSize) {

				vol->set(val[i], i);
				i++;
			}

		}
		);


		input.close();

		delete[] val;
	}
	else {
		std::cout << "Unable to open file";
		return false;
	}
	std::cout << "done reading" << std::endl;

	clock_t end = clock();
	double elapsed_secs = double(end - begin) / CLOCKS_PER_SEC;

	std::cout << "time: " << elapsed_secs << "s" << std::endl;


	return true;
}


void GLWidget::write(std::string file, VectorField* vf) {
	std::cout << "writing vector-field to file " << file << std::endl;

	std::ofstream filestream(file);
	if (filestream.is_open())
	{
		for (int i = 0; i < vf->size(); i++) {
			auto vec = vf->get(i);
			filestream << vec.x << " " << vec.y << " " << vec.z << "\n";
		}

		filestream.close();
	}
	else std::cout << "Unable to open file";

	std::cout << "done writing" << std::endl;
}

void GLWidget::write(std::string file, Volume* vol) {

	std::cout << "writing volume to file " << file << std::endl;

	std::ofstream filestream(file);
	if (filestream.is_open())
	{
		for (int i = 0; i < vol->size(); i++) {
			auto vox = vol->voxel(i);
			filestream << vox.getValue() << "\n";
		}

		filestream.close();
	}
	else std::cout << "Unable to open file";

	std::cout << "done writing" << std::endl;
}

bool GLWidget::renderSideBySide(int state)
{
	if (unfiltered == nullptr) {
		if (state == 2) {
			std::cout << "Error: Load volume fist" << std::endl;
		}
		return false;
	}

	if (state == 2) {

		sideBySide = true;
		m_camera.setAspect((float(width()) / 2.0f) / float(height()));


		if (streamline == nullptr) {
			streamline = streamlineAlgorithm();
		}

		destroyRecreateVolTexture(unfiltered, 1);
		destroyRecreateVolTexture(streamline, 2);

	}
	else {
		sideBySide = false;
		m_camera.setAspect(float(width()) / float(height()));

		filterChanged(QString(filter.c_str()));
	}

	return true;
}

bool GLWidget::zero(glm::vec3& v) {
	return v.x == 0 && v.y == 0 && v.z == 0;
}

void GLWidget::cleanup()
{
	makeCurrent();
	delete m_programCube;
	delete m_programVolume;
	doneCurrent();
}

void GLWidget::fileChanged(const QString &path)
{
	// reboot glsw, otherwise it will use the old cached shader
	glswShutdown();
	initglsw();

	loadShaders();
	update();
}

void GLWidget::loadShaders()
{
	bool success = false;

	const char *vs = glswGetShader("cube.Vertex");
	if (!vs)
	{
		qDebug() << "Error: not able to read vertex shader";
	}
	QOpenGLShader *vertexShader = new QOpenGLShader(QOpenGLShader::Vertex);
	success = vertexShader->compileSourceCode(vs);

	const char *fs = glswGetShader("cube.Fragment");
	if (!fs)
	{
		qDebug() << "Error: not able to read fragment shader";
	}
	QOpenGLShader *fragmentShader = new QOpenGLShader(QOpenGLShader::Fragment);
	success = fragmentShader->compileSourceCode(fs);

	m_programCube->removeAllShaders();
	m_programCube->addShader(vertexShader);
	m_programCube->addShader(fragmentShader);

	const char *vsVolume = glswGetShader("raycasting.Vertex");
	if (!vsVolume)
	{
		qDebug() << "Error: not able to read vertex shader";
	}
	QOpenGLShader *vertexShaderVolume = new QOpenGLShader(QOpenGLShader::Vertex);
	success = vertexShaderVolume->compileSourceCode(vsVolume);

	const char *fsVolume = glswGetShader("raycasting.Fragment");
	if (!fsVolume)
	{
		qDebug() << "Error: not able to read fragment shader";
	}
	QOpenGLShader *fragmentShaderVolume = new QOpenGLShader(QOpenGLShader::Fragment);
	success = fragmentShaderVolume->compileSourceCode(fsVolume);
	
	m_programVolume->removeAllShaders();
	m_programVolume->addShader(vertexShaderVolume);
	m_programVolume->addShader(fragmentShaderVolume);
	m_programVolume->link();
	
}

// this function is called whenever the widget has been resized
//resizing disabled
void GLWidget::resizeGL(int w, int h)
{

	// when the widget is resized, the aspect ratio of the camera needs to be updated
	m_camera.setAspect(float(w) / h);
}

void GLWidget::mousePressEvent(QMouseEvent *event)
{
	m_lastPos = event->pos();
}

void GLWidget::wheelEvent(QWheelEvent *event)
{
	//vertical scroll = zoom
	if (event->orientation() == Qt::Orientation::Vertical) {
		m_camera.update(0, 0, 0, event->delta() / 10.0f, 0, 0);

	}

	//horizontal scroll = roll
	if (event->orientation() == Qt::Orientation::Horizontal) {
		m_camera.update(0, 0, event->delta() / 1000.0f, 0, 0, 0);
	}

	update();
}

void GLWidget::mouseMoveEvent(QMouseEvent *event)
{
	int dx = event->x() - m_lastPos.x();
	int dy = event->y() - m_lastPos.y();

	//left mouse button = rotate camera
	if (event->buttons() & Qt::LeftButton) {
		m_camera.update(dy / 100.0f, dx / 100.0f, 0, 0, 0, 0);
	}

	//right mouse button = strafing
	if (event->buttons() & Qt::RightButton) {
		m_camera.update(0, 0, 0, 0, dx / 1000.0f, dy / 1000.0f);
	}

	m_lastPos = event->pos();
	update();
}

void GLWidget::keyPressEvent(QKeyEvent *event)
{
	switch (event->key())
	{
	case Qt::Key_Space:
	{
		break;
	}
	default:
	{
		event->ignore();
		break;
	}
	}
}

void GLWidget::keyReleaseEvent(QKeyEvent *event)
{
}

void GLWidget::handleLoggedMessage()
{
	const QList<QOpenGLDebugMessage> messages = m_logger->loggedMessages();
	for (const QOpenGLDebugMessage &message : messages)
	{
		qDebug() << message;
	}
}