/*
* Copyright (C) 2017
* Computer Graphics Group, The Institute of Computer Graphics and Algorithms, TU Wien
* Written by Tobias Klein <tklein@cg.tuwien.ac.at>
* All rights reserved.
*/

#pragma once

#include <QOpenGLWidget>
#include <QOpenGLFunctions>
#include <QOpenGLShaderProgram>
#include <QOpenGLVertexArrayObject>
#include <QOpenGLBuffer>
#include <QGLShader>
#include <QOpenGLShaderProgram>
#include <QFileSystemWatcher>
#include <QOpenGLDebugLogger>
#include <QOpenGLTexture>

#include <QTime>

#include "Camera.h"
#include "Volume.h"
#include "VectorField.h"

class MainWindow;

class GLWidget : public QOpenGLWidget, protected QOpenGLFunctions
{
	Q_OBJECT

public:
	GLWidget(QWidget *parent, MainWindow *mainWindow);
	~GLWidget();

	int _n = 5;
	int _m = 5;
	int _delta = 5;

	std::string filter = "Off";

	Camera m_camera;
	int m_renderingMode;

	/**
	generating all directions from a point P sampled by _delta steps
	stay the same for every point in every dataset, so it is generated only when _delta changes
	*/
	void generateDirs();

	void filterChanged(const QString & text = nullptr);

	/**
	generates basic gauss filtering of the current volume
	*/
	void calcGauss();

	/**
	starts the computation of the steamline algorithm and saves the output in streamline
	*/
	void calcStreamline();

	void destroyRecreateVolTexture(Volume * vol, int n = 1);

	/**
	Step 1: Local Direction of Lowest Variance
	*/
	VectorField* localDirectionOfLowestVariance();

	/**
	starts the whole streamline computation algorithm
	*/
	Volume * streamlineAlgorithm();

	/**
	Streamline Integration and Filtering
	*/
	Volume * streamlineIntegrationAndFiltering(VectorField * lowestVarDirs);

	/**
	reads vectorfield from file, if one with current settings is available
	*/
	bool read(std::ifstream & input, VectorField * lowestVarDirs, int volSize);

	/**
	reads already streamline filtered volume from file if available
	*/
	bool readFilteredVol(std::ifstream & input, Volume * vol, int volSize);

	/**
	writes vector field to file
	*/
	void write(std::string file, VectorField * lowestVarDirs);

	/**
	writes volume to file
	*/
	void write(std::string file, Volume * vol);

	/**
	called when en- or disabling side by side rendering
	*/
	bool renderSideBySide(int state);


public slots:
	void cleanup();
	void volumeTextureLoaded(Volume *v, std::string name);

signals:

protected:

	void mousePressEvent(QMouseEvent *event) Q_DECL_OVERRIDE;
	void mouseMoveEvent(QMouseEvent *event) Q_DECL_OVERRIDE;
	void keyPressEvent(QKeyEvent *event) Q_DECL_OVERRIDE;
	void keyReleaseEvent(QKeyEvent *event) Q_DECL_OVERRIDE;
	void wheelEvent(QWheelEvent *event) Q_DECL_OVERRIDE;

protected slots:

	void paintGL() Q_DECL_OVERRIDE;
	void initializeGL() Q_DECL_OVERRIDE;
	void resizeGL(int w, int h) Q_DECL_OVERRIDE;
	void fileChanged(const QString &path);
	void handleLoggedMessage();

private:

	bool sideBySide = false;

	std::string volumeName;

	/**
	checks if v is zero vector
	*/
	bool zero(glm::vec3& v);

	Volume* unfiltered = nullptr;
	Volume* streamline = nullptr;
	Volume* gauss = nullptr;

	float scaleX = 1;
	float scaleY = 1;
	float scaleZ = 1;


	void initglsw();
	void loadShaders();

	QOpenGLShaderProgram *m_programCube;
	QOpenGLShaderProgram *m_programVolume;
	
	QOpenGLShader *m_vertexShaderVolume;
	QOpenGLShader *m_fragmentShaderVolume;

	QOpenGLVertexArrayObject m_vaoCube;
	
	QOpenGLVertexArrayObject m_vaoQuad;
	
	QOpenGLBuffer m_vertexBuffer;
	QOpenGLBuffer m_indexBuffer;
	QOpenGLBuffer m_texCoordBuffer;

	QOpenGLBuffer m_vertexBufferQuad;
	QOpenGLBuffer m_indexBufferQuad;

	QOpenGLTexture *m_VolumeTexture;
	QOpenGLTexture *m_VolumeTexture2;

	QOpenGLDebugLogger *m_logger;
	
	QOpenGLFramebufferObject *m_FBO_frontFaces;
	QOpenGLFramebufferObject *m_FBO_backFaces;

	QFileSystemWatcher *m_fileWatcher;

	QPoint m_lastPos;

	MainWindow *m_MainWindow;

	std::vector<glm::vec3> dir_x;

};


