#include "VectorField.h"

VectorField::VectorField(int w, int h, int d)
	: sizeX(w), sizeY(h), sizeZ(d), _size(w*h*d)
{
	vectors = new glm::vec3[_size];
}

VectorField::~VectorField()
{
	delete[] vectors;
}

glm::vec3 VectorField::get(int x, int y, int z)
{
	if (x < 0 || y < 0 || z < 0 ||
		x >= sizeX || y >= sizeY || z >= sizeZ)
		return glm::vec3(0);

	return vectors[x + y * sizeX + z * sizeX * sizeY];
}

glm::vec3 VectorField::get(const glm::ivec3& p)
{
	return get(p.x, p.y, p.z);
}

glm::vec3 VectorField::get(int i)
{
	if (i < 0 || i >= _size)
		return glm::vec3(0);

	return vectors[i];
}

glm::vec3 VectorField::interpolate(const glm::vec3 & p)
{
	if (p.x < 0 || p.y < 0 || p.z < 0 ||
		p.x >= sizeX || p.y >= sizeY || p.z >= sizeZ) {
		return glm::vec3(0);
	}

	int ceil_x = int(ceil(p.x));
	int ceil_y = int(ceil(p.y));
	int ceil_z = int(ceil(p.z));

	int floor_x = int(floor(p.x));
	int floor_y = int(floor(p.y));
	int floor_z = int(floor(p.z));

	int cf_x = ceil_x - floor_x;
	int cf_y = ceil_y - floor_y;
	int cf_z = ceil_z - floor_z;

	if (cf_x == 0 &&
		cf_y == 0 &&
		cf_z == 0) {

		return get(int(p.x), int(p.y), int(p.z));
	}

	float x_d = cf_x == 0 ? 0 : (p.x - floor_x) / cf_x;
	float y_d = cf_y == 0 ? 0 : (p.y - floor_y) / cf_y;
	float z_d = cf_z == 0 ? 0 : (p.z - floor_z) / cf_z;


	// see https://en.wikipedia.org/wiki/Trilinear_interpolation

	auto c000 = get(floor_x, floor_y, floor_z);
	if (c000.x == 0 && c000.y == 0 && c000.z == 0) return glm::vec3(0);
	auto c001 = get(floor_x, floor_y, ceil_z);
	if (c001.x == 0 && c001.y == 0 && c001.z == 0) return glm::vec3(0);
	auto c010 = get(floor_x, ceil_y, floor_z);
	if (c010.x == 0 && c010.y == 0 && c010.z == 0) return glm::vec3(0);
	auto c011 = get(floor_x, ceil_y, ceil_z);
	if (c011.x == 0 && c011.y == 0 && c011.z == 0) return glm::vec3(0);

	auto c100 = get(ceil_x, floor_y, floor_z);
	if (c100.x == 0 && c100.y == 0 && c100.z == 0) return glm::vec3(0);
	auto c101 = get(ceil_x, floor_y, ceil_z);
	if (c101.x == 0 && c101.y == 0 && c101.z == 0) return glm::vec3(0);
	auto c110 = get(ceil_x, ceil_y, floor_z);
	if (c110.x == 0 && c110.y == 0 && c110.z == 0) return glm::vec3(0);
	auto c111 = get(ceil_x, ceil_y, ceil_z);
	if (c111.x == 0 && c111.y == 0 && c111.z == 0) return glm::vec3(0);

	auto c00 = c000 * (1 - x_d) + c100 * x_d;
	auto c01 = c001 * (1 - x_d) + c101 * x_d;
	auto c10 = c010 * (1 - x_d) + c110 * x_d;
	auto c11 = c011 * (1 - x_d) + c111 * x_d;

	auto c0 = c00 * (1 - y_d) + c10 * y_d;
	auto c1 = c01 * (1 - y_d) + c11 * y_d;

	auto c = c0 * (1 - z_d) + c1 * z_d;

	return c;
}

bool VectorField::set(glm::vec3 v, int x, int y, int z)
{
	if (x < 0 || y < 0 || z < 0 ||
		x >= sizeX || y >= sizeY || z >= sizeZ)
		return false;

	vectors[x + y * sizeX + z * sizeX * sizeY] = v;
	return true;
}

bool VectorField::set(glm::vec3 v, int i)
{
	if (i < 0 || i >= _size)
		return false;

	vectors[i] = v;
	return true;
}

int VectorField::size()
{
	return _size;
}
