#include "Volume.h"

#include <math.h>

#include <QDebug> 

#include <iostream>
#include <fstream>
#include <regex>
#include <ppl.h> 



//-------------------------------------------------------------------------------------------------
// Voxel
//-------------------------------------------------------------------------------------------------

Voxel::Voxel()
{
	setValue(0.0f);
}

Voxel::Voxel(const Voxel &other)
{
	setValue(other.getValue());
}

Voxel::Voxel(const float value)
{
	setValue(value);
}

Voxel::~Voxel()
{
}

void Voxel::setValue(const float value)
{
	m_Value = value;
}

const float Voxel::getValue() const
{
	return m_Value;
};

const bool Voxel::operator==(const Voxel &other) const
{
	return (getValue() == other.getValue());
};

const bool Voxel::operator!=(const Voxel &other) const
{
	return !(*this == other);
};

const bool Voxel::operator>(const Voxel &other) const
{
	return getValue() > other.getValue();
};

const bool Voxel::operator>=(const Voxel &other) const
{
	return getValue() >= other.getValue();
};

const bool Voxel::operator<(const Voxel &other) const
{
	return getValue() < other.getValue();
};

const bool Voxel::operator<=(const Voxel &other) const
{
	return getValue() <= other.getValue();
};

const Voxel& Voxel::operator+=(const Voxel &other)
{
	m_Value += other.m_Value;
	return *this;
};

const Voxel& Voxel::operator-=(const Voxel &other)
{
	m_Value -= other.m_Value;
	return *this;
};

const Voxel& Voxel::operator*=(const float &value)
{
	m_Value *= value;
	return *this;
};

const Voxel& Voxel::operator/=(const float &value)
{
	m_Value /= value;
	return *this;
};

const Voxel Voxel::operator+(const Voxel &other) const
{
	Voxel voxNew = *this;
	voxNew += other;
	return voxNew;
};

const Voxel Voxel::operator-(const Voxel &other) const
{
	Voxel voxNew = *this;
	voxNew -= other;
	return voxNew;
};

const Voxel Voxel::operator*(const float &value) const
{
	Voxel voxNew = *this;
	voxNew *= value;
	return voxNew;
};

const Voxel Voxel::operator/(const float &value) const
{
	Voxel voxNew = *this;
	voxNew /= value;
	return voxNew;
};


//-------------------------------------------------------------------------------------------------
// Volume
//-------------------------------------------------------------------------------------------------

Volume::Volume()
	: m_Width(1), m_Height(1), m_Depth(1), m_Size(0), m_Voxels(1)
{
}

Volume::Volume(int w, int h, int d)
	: m_Width(w), m_Height(h), m_Depth(d), m_Size(w*h*d), m_Voxels(1)
{
	m_Voxels.resize(m_Size, Voxel(0));
}

Volume::~Volume()
{
}

const Voxel Volume::voxel(const int x, const int y, const int z) const
{
	if (x < 0 || y < 0 || z < 0 ||
		x >= m_Width || y >= m_Height || z >= m_Depth) {
		return Voxel(-1);
	}

	return m_Voxels[x + y*m_Width + z*m_Width*m_Height];
}

const Voxel Volume::voxel(const glm::vec3& v) const
{
	return voxel(v.x, v.y, v.z);
}

const float Volume::voxelInterpolate(const glm::vec3& p) const
{
	return voxelInterpolate(p.x, p.y, p.z);
}

const float Volume::voxelInterpolate(const float x, const float y, const float z) const
{
	if (x < 0 || y < 0 || z < 0 ||
		x >= m_Width || y >= m_Height || z >= m_Depth) {
		return -1;
	}

	int ceil_x = ceil(x);
	int ceil_y = ceil(y);
	int ceil_z = ceil(z);

	int floor_x = floor(x);
	int floor_y = floor(y);
	int floor_z = floor(z);

	int cf_x = ceil_x - floor_x;
	int cf_y = ceil_y - floor_y;
	int cf_z = ceil_z - floor_z;

	if (cf_x == 0 &&
		cf_y == 0 &&
		cf_z == 0) {

		return voxel(x, y, z).getValue();
	}

	float x_d = cf_x == 0 ? 0 : (x - floor_x) / cf_x;
	float y_d = cf_y == 0 ? 0 : (y - floor_y) / cf_y;
	float z_d = cf_z == 0 ? 0 : (z - floor_z) / cf_z;


	// see https://en.wikipedia.org/wiki/Trilinear_interpolation

	auto c000 = voxel(floor_x, floor_y, floor_z).getValue();
	if (c000 < 0) return -1;
	auto c001 = voxel(floor_x, floor_y, ceil_z).getValue();
	if (c001 < 0) return -1;
	auto c010 = voxel(floor_x, ceil_y, floor_z).getValue();
	if (c010 < 0) return -1;
	auto c011 = voxel(floor_x, ceil_y, ceil_z).getValue();
	if (c011 < 0) return -1;

	auto c100 = voxel(ceil_x, floor_y, floor_z).getValue();
	if (c100 < 0) return -1;
	auto c101 = voxel(ceil_x, floor_y, ceil_z).getValue();
	if (c101 < 0) return -1;
	auto c110 = voxel(ceil_x, ceil_y, floor_z).getValue();
	if (c110 < 0) return -1;
	auto c111 = voxel(ceil_x, ceil_y, ceil_z).getValue();
	if (c111 < 0) return -1;

	auto c00 = c000 * (1 - x_d) + c100 * x_d;
	auto c01 = c001 * (1 - x_d) + c101 * x_d;
	auto c10 = c010 * (1 - x_d) + c110 * x_d;
	auto c11 = c011 * (1 - x_d) + c111 * x_d;

	auto c0 = c00 * (1 - y_d) + c10 * y_d;
	auto c1 = c01 * (1 - y_d) + c11 * y_d;

	auto c = c0 * (1 - z_d) + c1 * z_d;

	return c;
}

const Voxel& Volume::voxel(const int i) const
{
	return m_Voxels[i];
}

const Voxel* Volume::voxels() const
{
	return &(m_Voxels.front());
}

bool Volume::set(float f, int i) {

	if (i < 0 || i >= m_Size) {
		return false;
	}

	m_Voxels[i] = Voxel(f);
	return true;
}

bool Volume::set(float f, int x, int y, int z)
{
	if (x < 0 || y < 0 || z < 0 ||
		x >= m_Width || y >= m_Height || z >= m_Depth) {
		return false;
	}

	m_Voxels[x + y * m_Width + z * m_Width*m_Height] = Voxel(f);
	return true;
}
;

const int Volume::width() const
{
	return m_Width;
};

const int Volume::height() const
{
	return m_Height;
};

const int Volume::depth() const
{
	return m_Depth;
};

const int Volume::size() const
{
	return m_Size;
};


//-------------------------------------------------------------------------------------------------
// Volume File Loader
//-------------------------------------------------------------------------------------------------

bool Volume::loadFromFile(QString filename, QProgressBar* progressBar, std::string type)
{
	FILE *fp = NULL;
	std::ifstream input;

	// load file
	fopen_s(&fp, filename.toStdString().c_str(), "rb");
	if (!fp)
	{
		std::cerr << "+ Error loading file: " << filename.toStdString() << std::endl;
		return false;
	}


	if (type == "dat") {

		// read header and set volume dimensions

		unsigned short uWidth, uHeight, uDepth;
		fread(&uWidth, sizeof(unsigned short), 1, fp);
		fread(&uHeight, sizeof(unsigned short), 1, fp);
		fread(&uDepth, sizeof(unsigned short), 1, fp);

		m_Width = int(uWidth);
		m_Height = int(uHeight);
		m_Depth = int(uDepth);

	}
	else if (type == "raw") {

		//no header, get dimensions from filename

		std::string fn = filename.toStdString();

		std::string::size_type w_start = fn.find('(');
		std::string::size_type w_end = fn.find('x', w_start + 2);
		m_Width = std::stoi(fn.substr(w_start + 1, w_end - w_start - 1));

		std::string::size_type h_start = fn.find('x');
		std::string::size_type h_end = fn.find_last_of('x', h_start + 2);
		m_Height = std::stoi(fn.substr(h_start + 1, h_end - h_start - 1));

		std::string::size_type d_start = fn.find_last_of('x');
		std::string::size_type d_end = fn.find(')', d_start + 2);
		m_Depth = std::stoi(fn.substr(d_start + 1, d_end - d_start - 1));
	}

	int slice = m_Width * m_Height;
	m_Size = slice * m_Depth;

	m_Voxels.resize(m_Size);

	// progress bar

	progressBar->setRange(0, m_Size + 10);
	progressBar->setValue(0);


	// read volume data

	if (type == "dat") {
		// read into vector before writing data into volume to speed up process
		std::vector<unsigned short> vecData;
		vecData.resize(m_Size);
		fread((void*)&(vecData.front()), sizeof(unsigned short), m_Size, fp);
		fclose(fp);


		// store volume data

		int digits = 0;
		while (pow(10, digits) < m_Size) { digits++; }
		int blocksize = pow(10, (digits - 2));

		concurrency::parallel_for(int(0), m_Size / blocksize, [&](int block) {

			int i = block * blocksize;

			while (i < (block + 1) * blocksize  && i < m_Size) {
				// data is converted to FLOAT values in an interval of [0.0 .. 1.0];
				const float value = float(vecData[i]) / 4095.0f;
				m_Voxels[i] = Voxel(value);

				i++;
			}

		}
		);


	}
	else if (type == "raw") {

		std::vector<unsigned char> buffer;
		buffer.resize(m_Size);
		fread((void*)&(buffer.front()), sizeof(unsigned char), m_Size, fp);
		fclose(fp);

		// store volume data

		int digits = 0;
		while (pow(10,digits) < m_Size) { digits++; }
		int blocksize = pow(10, (digits - 2));
		
		concurrency::parallel_for(int(0), m_Size/blocksize, [&](int block) {

			int i = block * blocksize;

			while (i < (block + 1) * blocksize  && i < m_Size) {
				// data is converted to FLOAT values in an interval of [0.0 .. 1.0];
				const float value = float(buffer[i]) / 255.0f;
				m_Voxels[i] = Voxel(value);

				i++;
			}

		}
		);

	}
	

	progressBar->setValue(0);

	qDebug() << "Loaded VOLUME with dimensions " << QString::number(m_Width) << " x " << QString::number(m_Height) << " x " << QString::number(m_Depth);

	return true;
}